package javax.swing;

import java.security.PrivilegedAction;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import java.util.Iterator;
import java.util.ListIterator;
import java.lang.reflect.InvocationTargetException;
import java.util.Collections;
import java.util.ArrayList;
import java.lang.reflect.Method;
import java.util.List;
import java.awt.Container;
import sun.util.logging.PlatformLogger;
import java.awt.Component;
import java.util.Comparator;

public class SortingFocusTraversalPolicy extends InternalFrameFocusTraversalPolicy
{
    private Comparator<? super Component> comparator;
    private boolean implicitDownCycleTraversal;
    private PlatformLogger log;
    private transient Container cachedRoot;
    private transient List<Component> cachedCycle;
    private static final SwingContainerOrderFocusTraversalPolicy fitnessTestPolicy;
    private final int FORWARD_TRAVERSAL = 0;
    private final int BACKWARD_TRAVERSAL = 1;
    private static final boolean legacySortingFTPEnabled;
    private static final Method legacyMergeSortMethod;
    
    protected SortingFocusTraversalPolicy() {
        this.implicitDownCycleTraversal = true;
        this.log = PlatformLogger.getLogger("javax.swing.SortingFocusTraversalPolicy");
    }
    
    public SortingFocusTraversalPolicy(Comparator<? super Component> comparator) {
        this.implicitDownCycleTraversal = true;
        this.log = PlatformLogger.getLogger("javax.swing.SortingFocusTraversalPolicy");
        this.comparator = comparator;
    }
    
    private List<Component> getFocusTraversalCycle(Container container) {
        ArrayList<Component> list = new ArrayList();
        this.enumerateAndSortCycle(container, list);
        return list;
    }
    
    private int getComponentIndex(List<Component> list, Component component) {
        int n;
        try {
            n = Collections.binarySearch(list, component, this.comparator);
        }
        catch (ClassCastException t) {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### During the binary search for " + component + " the exception occurred: ", t);
            }
            return -1;
        }
        if (n < 0) {
            n = list.indexOf(component);
        }
        return n;
    }
    
    private void enumerateAndSortCycle(Container container, List<Component> list) {
        if (container.isShowing()) {
            this.enumerateCycle(container, list);
            if (!SortingFocusTraversalPolicy.legacySortingFTPEnabled || !this.legacySort(list, this.comparator)) {
                Collections.sort(list, this.comparator);
            }
        }
    }
    
    private boolean legacySort(List<Component> list, Comparator<? super Component> comparator) {
        if (SortingFocusTraversalPolicy.legacyMergeSortMethod == null) {
            return false;
        }
        Object[] array = list.toArray();
        try {
            SortingFocusTraversalPolicy.legacyMergeSortMethod.invoke(null, new Object[] { array, comparator });
        }
        catch (IllegalAccessException | InvocationTargetException ex) {
            return false;
        }
        ListIterator<Component> listIterator = list.listIterator();
        for (Object o : array) {
            listIterator.next();
            listIterator.set((Component)o);
        }
        return true;
    }
    
    private void enumerateCycle(Container container, List<Component> list) {
        if (!container.isVisible() || !container.isDisplayable()) {
            return;
        }
        list.add(container);
        Component[] components;
        for (Component component : components = container.getComponents()) {
            if ((component instanceof Container)) {
                Container container2 = (Container)component;
                if (!container2.isFocusCycleRoot() && !container2.isFocusTraversalPolicyProvider() && (!(container2 instanceof JComponent) || !((JComponent)container2).isManagingFocus())) {
                    this.enumerateCycle(container2, list);
                    continue;
                }
            }
            list.add(component);
        }
        return;
    }
    
    Container getTopmostProvider(Container container, Component component) {
        Container container2 = component.getParent();
        Container container3 = null;
        while (container2 != container && container2 != null) {
            if (container2.isFocusTraversalPolicyProvider()) {
                container3 = container2;
            }
            container2 = container2.getParent();
        }
        if (container2 == null) {
            return null;
        }
        return container3;
    }
    
    private Component getComponentDownCycle(Component component, int n) {
        Component defaultComponent = null;
        if ((component instanceof Container)) {
            Container container = (Container)component;
            if (container.isFocusCycleRoot()) {
                if (this.getImplicitDownCycleTraversal()) {
                    defaultComponent = container.getFocusTraversalPolicy().getDefaultComponent(container);
                    if (defaultComponent != null && this.log.isLoggable(PlatformLogger.Level.FINE)) {
                        this.log.fine("### Transfered focus down-cycle to " + defaultComponent + " in the focus cycle root " + container);
                    }
                }
                else {
                    return null;
                }
            }
            else if (container.isFocusTraversalPolicyProvider()) {
                defaultComponent = ((n == 0) ? container.getFocusTraversalPolicy().getDefaultComponent(container) : container.getFocusTraversalPolicy().getLastComponent(container));
                if (defaultComponent != null && this.log.isLoggable(PlatformLogger.Level.FINE)) {
                    this.log.fine("### Transfered focus to " + defaultComponent + " in the FTP provider " + container);
                }
            }
        }
        return defaultComponent;
    }
    
    @Override
    public Component getComponentAfter(Container container, Component component) {
        if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
            this.log.fine("### Searching in " + container + " for component after " + component);
        }
        if (container == null || component == null) {
            throw new IllegalArgumentException("aContainer and aComponent cannot be null");
        }
        if (!container.isFocusTraversalPolicyProvider() && !container.isFocusCycleRoot()) {
            throw new IllegalArgumentException("aContainer should be focus cycle root or focus traversal policy provider");
        }
        if (container.isFocusCycleRoot() && !component.isFocusCycleRoot(container)) {
            throw new IllegalArgumentException("aContainer is not a focus cycle root of aComponent");
        }
        Component componentDownCycle = this.getComponentDownCycle(component, 0);
        if (componentDownCycle != null) {
            return componentDownCycle;
        }
        Container topmostProvider = this.getTopmostProvider(container, component);
        if (topmostProvider != null) {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Asking FTP " + topmostProvider + " for component after " + component);
            }
            Component componentAfter = topmostProvider.getFocusTraversalPolicy().getComponentAfter(topmostProvider, component);
            if (componentAfter != null) {
                if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                    this.log.fine("### FTP returned " + componentAfter);
                    return componentAfter;
                }
                return componentAfter;
            }
            else {
                component = topmostProvider;
            }
        }
        List<Component> focusTraversalCycle = this.getFocusTraversalCycle(container);
        if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
            this.log.fine("### Cycle is " + focusTraversalCycle + ", component is " + component);
        }
        int i = this.getComponentIndex(focusTraversalCycle, component);
        if (i < 0) {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Didn't find component " + component + " in a cycle " + container);
            }
            return this.getFirstComponent(container);
        }
        else {
            ++i;
            while (i < focusTraversalCycle.size()) {
                Component component2 = (Component)focusTraversalCycle.get(i);
                if (this.accept(component2)) {
                    return component2;
                }
                Component componentDownCycle2;
                if ((componentDownCycle2 = this.getComponentDownCycle(component2, 0)) != null) {
                    return componentDownCycle2;
                }
                ++i;
            }
            if (container.isFocusCycleRoot()) {
                this.cachedRoot = container;
                this.cachedCycle = focusTraversalCycle;
                Component firstComponent = this.getFirstComponent(container);
                this.cachedRoot = null;
                this.cachedCycle = null;
                return firstComponent;
            }
            return null;
        }
    }
    
    @Override
    public Component getComponentBefore(Container obj, Component obj2) {
        if (obj == null || obj2 == null) {
            throw new IllegalArgumentException("aContainer and aComponent cannot be null");
        }
        if (!obj.isFocusTraversalPolicyProvider() && !obj.isFocusCycleRoot()) {
            throw new IllegalArgumentException("aContainer should be focus cycle root or focus traversal policy provider");
        }
        if (obj.isFocusCycleRoot() && !obj2.isFocusCycleRoot(obj)) {
            throw new IllegalArgumentException("aContainer is not a focus cycle root of aComponent");
        }
        Container topmostProvider = this.getTopmostProvider(obj, obj2);
        if (topmostProvider != null) {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Asking FTP " + topmostProvider + " for component after " + obj2);
            }
            Component componentBefore = topmostProvider.getFocusTraversalPolicy().getComponentBefore(topmostProvider, obj2);
            if (componentBefore != null) {
                if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                    this.log.fine("### FTP returned " + componentBefore);
                    return componentBefore;
                }
                return componentBefore;
            }
            else {
                obj2 = topmostProvider;
                if (this.accept(obj2)) {
                    return obj2;
                }
            }
        }
        List<Component> focusTraversalCycle = this.getFocusTraversalCycle(obj);
        if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
            this.log.fine("### Cycle is " + focusTraversalCycle + ", component is " + obj2);
        }
        int i = this.getComponentIndex(focusTraversalCycle, obj2);
        if (i < 0) {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Didn't find component " + obj2 + " in a cycle " + obj);
            }
            return this.getLastComponent(obj);
        }
        else {
            --i;
            while (i >= 0) {
                Component component = (Component)focusTraversalCycle.get(i);
                Component componentDownCycle;
                if (component != obj && (componentDownCycle = this.getComponentDownCycle(component, 1)) != null) {
                    return componentDownCycle;
                }
                if (this.accept(component)) {
                    return component;
                }
                --i;
            }
            if (obj.isFocusCycleRoot()) {
                this.cachedRoot = obj;
                this.cachedCycle = focusTraversalCycle;
                Component lastComponent = this.getLastComponent(obj);
                this.cachedRoot = null;
                this.cachedCycle = null;
                return lastComponent;
            }
            return null;
        }
    }
    
    @Override
    public Component getFirstComponent(Container obj) {
        if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
            this.log.fine("### Getting first component in " + obj);
        }
        if (obj == null) {
            throw new IllegalArgumentException("aContainer cannot be null");
        }
        List<Component> obj2;
        if (this.cachedRoot == obj) {
            obj2 = this.cachedCycle;
        }
        else {
            obj2 = this.getFocusTraversalCycle(obj);
        }
        if (obj2.size() == 0) {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Cycle is empty");
            }
            return null;
        }
        else {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Cycle is " + obj2);
            }
            for (Component component : obj2) {
                if (this.accept(component)) {
                    return component;
                }
                Component componentDownCycle;
                if (component != obj && (componentDownCycle = this.getComponentDownCycle(component, 0)) != null) {
                    return componentDownCycle;
                }
                continue;
            }
            return null;
        }
    }
    
    @Override
    public Component getLastComponent(Container obj) {
        if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
            this.log.fine("### Getting last component in " + obj);
        }
        if (obj == null) {
            throw new IllegalArgumentException("aContainer cannot be null");
        }
        List<Component> obj2;
        if (this.cachedRoot == obj) {
            obj2 = this.cachedCycle;
        }
        else {
            obj2 = this.getFocusTraversalCycle(obj);
        }
        if (obj2.size() == 0) {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Cycle is empty");
            }
            return null;
        }
        else {
            if (this.log.isLoggable(PlatformLogger.Level.FINE)) {
                this.log.fine("### Cycle is " + obj2);
            }
            int i = obj2.size() - 1;
            while (i >= 0) {
                Component component = (Component)obj2.get(i);
                if (this.accept(component)) {
                    return component;
                }
                if (component instanceof Container && component != obj) {
                    Container container = (Container)component;
                    if (container.isFocusTraversalPolicyProvider()) {
                        Component lastComponent = container.getFocusTraversalPolicy().getLastComponent(container);
                        if (lastComponent != null) {
                            return lastComponent;
                        }
                    }
                }
                --i;
            }
            return null;
        }
    }
    
    @Override
    public Component getDefaultComponent(Container container) {
        return this.getFirstComponent(container);
    }
    
    public void setImplicitDownCycleTraversal(boolean implicitDownCycleTraversal) {
        this.implicitDownCycleTraversal = implicitDownCycleTraversal;
    }
    
    public boolean getImplicitDownCycleTraversal() {
        return this.implicitDownCycleTraversal;
    }
    
    protected void setComparator(Comparator<? super Component> comparator) {
        this.comparator = comparator;
    }
    
    protected Comparator<? super Component> getComparator() {
        return this.comparator;
    }
    
    protected boolean accept(Component component) {
        return SortingFocusTraversalPolicy.fitnessTestPolicy.accept(component);
    }
    
    static {
        fitnessTestPolicy = new SwingContainerOrderFocusTraversalPolicy();
        legacySortingFTPEnabled = "true".equals(AccessController.doPrivileged(new GetPropertyAction("swing.legacySortingFTPEnabled", "true")));
        legacyMergeSortMethod = (SortingFocusTraversalPolicy.legacySortingFTPEnabled ? ((Method)AccessController.doPrivileged(new PrivilegedAction<Method>() {
            @Override
            public Method run() {
                try {
                    Method declaredMethod = Class.forName("java.util.Arrays").getDeclaredMethod("legacyMergeSort", new Class[] { Object[].class, Comparator.class });
                    declaredMethod.setAccessible(true);
                    return declaredMethod;
                }
                catch (ClassNotFoundException | NoSuchMethodException ex) {
                    return null;
                }
            }
        })) : null);
    }
}
